typedef float GLfloat;
typedef unsigned long GLenum;
typedef unsigned char GLubyte;
typedef unsigned short GLushort;
typedef unsigned long GLuint;


#define GL_UNSIGNED_BYTE                  0x1401
#define GL_UNSIGNED_SHORT                 0x1403
#define GL_UNSIGNED_INT                   0x1405
